# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from galts_trade_api.transport.protos import common_pb2 as protos_dot_common__pb2
from galts_trade_api.transport.protos.exchange_info import api_pb2 as protos_dot_exchange__info_dot_api__pb2
from galts_trade_api.transport.protos.exchange_info import entities_pb2 as protos_dot_exchange__info_dot_entities__pb2


class ExchangeInfoApiStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.StartSession = channel.unary_unary(
        '/ExchangeInfoApi/StartSession',
        request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        response_deserializer=protos_dot_common__pb2.SessionHash.FromString,
        )
    self.FinishSession = channel.unary_unary(
        '/ExchangeInfoApi/FinishSession',
        request_serializer=protos_dot_common__pb2.SessionHash.SerializeToString,
        response_deserializer=protos_dot_common__pb2.BoolResult.FromString,
        )
    self.SubscribeToEntitiesChange = channel.unary_stream(
        '/ExchangeInfoApi/SubscribeToEntitiesChange',
        request_serializer=protos_dot_exchange__info_dot_api__pb2.SubscribeToEntitiesChangeIn.SerializeToString,
        response_deserializer=protos_dot_exchange__info_dot_api__pb2.SubscribeToEntitiesChangeOut.FromString,
        )
    self.GetEntities = channel.unary_unary(
        '/ExchangeInfoApi/GetEntities',
        request_serializer=protos_dot_exchange__info_dot_api__pb2.EntityKinds.SerializeToString,
        response_deserializer=protos_dot_exchange__info_dot_api__pb2.Entities.FromString,
        )
    self.AddAsset = channel.unary_unary(
        '/ExchangeInfoApi/AddAsset',
        request_serializer=protos_dot_exchange__info_dot_api__pb2.AddAssetIn.SerializeToString,
        response_deserializer=protos_dot_exchange__info_dot_api__pb2.Id.FromString,
        )
    self.GetAsset = channel.unary_unary(
        '/ExchangeInfoApi/GetAsset',
        request_serializer=protos_dot_exchange__info_dot_api__pb2.Id.SerializeToString,
        response_deserializer=protos_dot_exchange__info_dot_entities__pb2.Asset.FromString,
        )
    self.DeleteAsset = channel.unary_unary(
        '/ExchangeInfoApi/DeleteAsset',
        request_serializer=protos_dot_exchange__info_dot_api__pb2.Id.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )
    self.AddSymbol = channel.unary_unary(
        '/ExchangeInfoApi/AddSymbol',
        request_serializer=protos_dot_exchange__info_dot_api__pb2.AddSymbolIn.SerializeToString,
        response_deserializer=protos_dot_exchange__info_dot_api__pb2.Id.FromString,
        )
    self.GetSymbol = channel.unary_unary(
        '/ExchangeInfoApi/GetSymbol',
        request_serializer=protos_dot_exchange__info_dot_api__pb2.Id.SerializeToString,
        response_deserializer=protos_dot_exchange__info_dot_entities__pb2.Symbol.FromString,
        )
    self.DeleteSymbol = channel.unary_unary(
        '/ExchangeInfoApi/DeleteSymbol',
        request_serializer=protos_dot_exchange__info_dot_api__pb2.Id.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )
    self.AddMarket = channel.unary_unary(
        '/ExchangeInfoApi/AddMarket',
        request_serializer=protos_dot_exchange__info_dot_api__pb2.AddMarketIn.SerializeToString,
        response_deserializer=protos_dot_exchange__info_dot_api__pb2.Id.FromString,
        )
    self.GetMarket = channel.unary_unary(
        '/ExchangeInfoApi/GetMarket',
        request_serializer=protos_dot_exchange__info_dot_api__pb2.Id.SerializeToString,
        response_deserializer=protos_dot_exchange__info_dot_entities__pb2.Market.FromString,
        )
    self.DeleteMarket = channel.unary_unary(
        '/ExchangeInfoApi/DeleteMarket',
        request_serializer=protos_dot_exchange__info_dot_api__pb2.Id.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )
    self.AddExchange = channel.unary_unary(
        '/ExchangeInfoApi/AddExchange',
        request_serializer=protos_dot_exchange__info_dot_api__pb2.AddExchangeIn.SerializeToString,
        response_deserializer=protos_dot_exchange__info_dot_api__pb2.Id.FromString,
        )
    self.GetExchange = channel.unary_unary(
        '/ExchangeInfoApi/GetExchange',
        request_serializer=protos_dot_exchange__info_dot_api__pb2.Id.SerializeToString,
        response_deserializer=protos_dot_exchange__info_dot_entities__pb2.Exchange.FromString,
        )
    self.DeleteExchange = channel.unary_unary(
        '/ExchangeInfoApi/DeleteExchange',
        request_serializer=protos_dot_exchange__info_dot_api__pb2.Id.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )
    self.DisableExchange = channel.unary_unary(
        '/ExchangeInfoApi/DisableExchange',
        request_serializer=protos_dot_exchange__info_dot_api__pb2.Id.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )
    self.EnableExchange = channel.unary_unary(
        '/ExchangeInfoApi/EnableExchange',
        request_serializer=protos_dot_exchange__info_dot_api__pb2.Id.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )


class ExchangeInfoApiServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def StartSession(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def FinishSession(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SubscribeToEntitiesChange(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetEntities(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def AddAsset(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAsset(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def DeleteAsset(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def AddSymbol(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetSymbol(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def DeleteSymbol(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def AddMarket(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetMarket(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def DeleteMarket(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def AddExchange(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetExchange(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def DeleteExchange(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def DisableExchange(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def EnableExchange(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_ExchangeInfoApiServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'StartSession': grpc.unary_unary_rpc_method_handler(
          servicer.StartSession,
          request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
          response_serializer=protos_dot_common__pb2.SessionHash.SerializeToString,
      ),
      'FinishSession': grpc.unary_unary_rpc_method_handler(
          servicer.FinishSession,
          request_deserializer=protos_dot_common__pb2.SessionHash.FromString,
          response_serializer=protos_dot_common__pb2.BoolResult.SerializeToString,
      ),
      'SubscribeToEntitiesChange': grpc.unary_stream_rpc_method_handler(
          servicer.SubscribeToEntitiesChange,
          request_deserializer=protos_dot_exchange__info_dot_api__pb2.SubscribeToEntitiesChangeIn.FromString,
          response_serializer=protos_dot_exchange__info_dot_api__pb2.SubscribeToEntitiesChangeOut.SerializeToString,
      ),
      'GetEntities': grpc.unary_unary_rpc_method_handler(
          servicer.GetEntities,
          request_deserializer=protos_dot_exchange__info_dot_api__pb2.EntityKinds.FromString,
          response_serializer=protos_dot_exchange__info_dot_api__pb2.Entities.SerializeToString,
      ),
      'AddAsset': grpc.unary_unary_rpc_method_handler(
          servicer.AddAsset,
          request_deserializer=protos_dot_exchange__info_dot_api__pb2.AddAssetIn.FromString,
          response_serializer=protos_dot_exchange__info_dot_api__pb2.Id.SerializeToString,
      ),
      'GetAsset': grpc.unary_unary_rpc_method_handler(
          servicer.GetAsset,
          request_deserializer=protos_dot_exchange__info_dot_api__pb2.Id.FromString,
          response_serializer=protos_dot_exchange__info_dot_entities__pb2.Asset.SerializeToString,
      ),
      'DeleteAsset': grpc.unary_unary_rpc_method_handler(
          servicer.DeleteAsset,
          request_deserializer=protos_dot_exchange__info_dot_api__pb2.Id.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
      'AddSymbol': grpc.unary_unary_rpc_method_handler(
          servicer.AddSymbol,
          request_deserializer=protos_dot_exchange__info_dot_api__pb2.AddSymbolIn.FromString,
          response_serializer=protos_dot_exchange__info_dot_api__pb2.Id.SerializeToString,
      ),
      'GetSymbol': grpc.unary_unary_rpc_method_handler(
          servicer.GetSymbol,
          request_deserializer=protos_dot_exchange__info_dot_api__pb2.Id.FromString,
          response_serializer=protos_dot_exchange__info_dot_entities__pb2.Symbol.SerializeToString,
      ),
      'DeleteSymbol': grpc.unary_unary_rpc_method_handler(
          servicer.DeleteSymbol,
          request_deserializer=protos_dot_exchange__info_dot_api__pb2.Id.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
      'AddMarket': grpc.unary_unary_rpc_method_handler(
          servicer.AddMarket,
          request_deserializer=protos_dot_exchange__info_dot_api__pb2.AddMarketIn.FromString,
          response_serializer=protos_dot_exchange__info_dot_api__pb2.Id.SerializeToString,
      ),
      'GetMarket': grpc.unary_unary_rpc_method_handler(
          servicer.GetMarket,
          request_deserializer=protos_dot_exchange__info_dot_api__pb2.Id.FromString,
          response_serializer=protos_dot_exchange__info_dot_entities__pb2.Market.SerializeToString,
      ),
      'DeleteMarket': grpc.unary_unary_rpc_method_handler(
          servicer.DeleteMarket,
          request_deserializer=protos_dot_exchange__info_dot_api__pb2.Id.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
      'AddExchange': grpc.unary_unary_rpc_method_handler(
          servicer.AddExchange,
          request_deserializer=protos_dot_exchange__info_dot_api__pb2.AddExchangeIn.FromString,
          response_serializer=protos_dot_exchange__info_dot_api__pb2.Id.SerializeToString,
      ),
      'GetExchange': grpc.unary_unary_rpc_method_handler(
          servicer.GetExchange,
          request_deserializer=protos_dot_exchange__info_dot_api__pb2.Id.FromString,
          response_serializer=protos_dot_exchange__info_dot_entities__pb2.Exchange.SerializeToString,
      ),
      'DeleteExchange': grpc.unary_unary_rpc_method_handler(
          servicer.DeleteExchange,
          request_deserializer=protos_dot_exchange__info_dot_api__pb2.Id.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
      'DisableExchange': grpc.unary_unary_rpc_method_handler(
          servicer.DisableExchange,
          request_deserializer=protos_dot_exchange__info_dot_api__pb2.Id.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
      'EnableExchange': grpc.unary_unary_rpc_method_handler(
          servicer.EnableExchange,
          request_deserializer=protos_dot_exchange__info_dot_api__pb2.Id.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'ExchangeInfoApi', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



import pickle
from typing import Dict, Tuple


class Singleton(type):
    _instances = {}
    _hash_args = []
    _hash_kwargs = {}

    def __new__(mcs, *args, singleton_hash_args=None, singleton_hash_kwargs=None, **kwargs):
        mcs._hash_args = singleton_hash_args if singleton_hash_args is not None else []
        mcs._hash_kwargs = singleton_hash_kwargs if singleton_hash_kwargs is not None else {}

        return super().__new__(mcs, *args, **kwargs)

    def __call__(cls, *args, **kwargs):
        args_hash = cls._calculate_args_hash(args, kwargs)
        key = (cls, args_hash)

        if key not in cls._instances:
            cls._instances[key] = super().__call__(*args, **kwargs)

        return cls._instances[key]

    @classmethod
    def _calculate_args_hash(mcs, args: Tuple, kwargs: Dict) -> int:
        filtered_args = [args[i] for i in range(len(args)) if i in mcs._hash_args]
        filtered_kwargs = {k: v for k, v in kwargs.items() if k in mcs._hash_kwargs}
        hash_source = (filtered_args, filtered_kwargs)

        result = hash(pickle.dumps(hash_source, protocol=pickle.HIGHEST_PROTOCOL))

        return result



from __future__ import annotations

import datetime
from typing import Dict, Optional

from .transport import TransportFactory


class Exchange:
    def __init__(
        self,
        transport_factory: TransportFactory,
        id: int,
        tag: str,
        name: str,
        create_time: datetime.datetime,
        delete_time: Optional[datetime.datetime],
        disable_time: Optional[datetime.datetime]
    ):
        self._transport_factory = transport_factory
        self._id: int = int(id)
        self._tag: str = str(tag).strip()
        self._name: str = str(name).strip()
        self._create_time: datetime.datetime = create_time
        self._delete_time: Optional[datetime.datetime] = delete_time
        self._disable_time: Optional[datetime.datetime] = disable_time

        self._markets: Dict[str, Market] = {}

    @property
    def transport_factory(self):
        return self._transport_factory

    @transport_factory.setter
    def transport_factory(self, value):
        self._transport_factory = value

    @property
    def tag(self):
        return self._tag

    def add_market(self, market: Market) -> None:
        if market.custom_tag in self._markets:
            raise ValueError(
                f'Market with tag "{market.custom_tag}" already exists in exchange {self.tag}'
            )

        self._markets[market.custom_tag] = market

    def get_market_by_custom_tag(self, custom_tag: str) -> Market:
        return self._markets[custom_tag]


class Market:
    def __init__(
        self,
        transport_factory: TransportFactory,
        id: int,
        custom_tag: str,
        exchange_id: int,
        symbol_id: int,
        trade_endpoint: str,
        create_time: datetime.datetime,
        delete_time: Optional[datetime.datetime]
    ):
        self._transport_factory = transport_factory
        self._id: int = int(id)
        self._custom_tag: str = str(custom_tag).strip()
        self._exchange_id: int = int(exchange_id)
        self._symbol_id: int = int(symbol_id)
        self._trade_endpoint: str = str(trade_endpoint).strip()
        self._create_time: datetime.datetime = create_time
        self._delete_time: Optional[datetime.datetime] = delete_time

    @property
    def transport_factory(self):
        return self._transport_factory

    @transport_factory.setter
    def transport_factory(self, value):
        self._transport_factory = value

    @property
    def custom_tag(self):
        return self._custom_tag



import datetime
from asyncio import Event, wait_for
from typing import Awaitable, Callable, Dict, List, MutableMapping, Optional

from .asset import Asset, Symbol
from .exchange import Exchange, Market
from .transport import DepthConsumeKey, TransportFactory

OnPriceCallable = Callable[[str, str, str, datetime.datetime, List, List], Awaitable]


class Terminal:
    def __init__(self, transport: TransportFactory):
        self._transport_factory: TransportFactory = transport
        self._exchange_entities_inited = Event()
        self._exchanges: Dict[str, Exchange] = {}
        self._assets: Dict[str, Asset] = {}
        self._symbols: Dict[int, Symbol] = {}

    @property
    def transport_factory(self):
        return self._transport_factory

    @transport_factory.setter
    def transport_factory(self, value):
        self._transport_factory = value

    async def init_transport(self, loop_debug: Optional[bool] = None) -> None:
        await self.transport_factory.init(loop_debug)

    def shutdown_transport(self) -> None:
        self.transport_factory.shutdown()

    def is_exchange_entities_inited(self) -> bool:
        return self._exchange_entities_inited.is_set()

    async def wait_exchange_entities_inited(self, timeout: float = 5.0) -> None:
        await wait_for(self._exchange_entities_inited.wait(), timeout)

    async def init_exchange_entities(self) -> None:
        await self.transport_factory.init_exchange_entities(
            self._on_init_exchange_entities_response
        )

    async def auth_user(self, username: str, password: str) -> bool:
        return True

    def get_exchange(self, tag: str) -> Exchange:
        return self._exchanges[tag]

    async def subscribe_to_prices(
        self,
        callback: OnPriceCallable,
        consume_keys: Optional[List[DepthConsumeKey]] = None
    ) -> None:
        await self.transport_factory.get_depth_scraping_consumer(
            lambda event: callback(*event),
            consume_keys
        )

    async def _on_init_exchange_entities_response(
        self,
        data: MutableMapping[str, MutableMapping]
    ) -> None:
        properties_to_fill = ('exchanges', 'markets', 'symbols', 'assets')

        for prop in properties_to_fill:
            if prop not in data:
                # @TODO Stop the process on this exception
                raise KeyError(f'init_exchange_entities data have not required key "{prop}"')

            data[prop] = {k: v for k, v in data[prop].items() if not v['delete_time']}

        for entity in data['assets'].values():
            key = entity['tag']
            if key in self._assets:
                raise ValueError(f'Asset with tag "{key}" already exists')

            self._assets[key] = Asset(self.transport_factory, **entity)

        for id_, entity in data['symbols'].items():
            if id_ in self._symbols:
                raise ValueError(f'Symbol with id {id_} already exists')

            self._symbols[id_] = Symbol(self.transport_factory, **entity)

        exchanges_ids_map = {}
        for entity in data['exchanges'].values():
            key = entity['tag']
            if key in self._exchanges:
                raise ValueError(f'Exchange with tag "{key}" already exists')

            exchange = Exchange(self.transport_factory, **entity)
            self._exchanges[key] = exchange
            exchanges_ids_map[entity['id']] = exchange

        for entity in data['markets'].values():
            key = entity['exchange_id']
            if key not in exchanges_ids_map:
                raise ValueError(
                    f'No exchange with id {key} has been found for market with id {entity["id"]}'
                )

            exchanges_ids_map[key].add_market(Market(self.transport_factory, **entity))

        self._exchange_entities_inited.set()



import uuid
from datetime import timezone
from decimal import Decimal
from typing import Any, Dict, List, Mapping, Optional, Union

from google.protobuf.descriptor import FieldDescriptor
from google.protobuf.message import Message
from google.protobuf.pyext._message import MessageMapContainer, RepeatedCompositeContainer, \
    RepeatedScalarContainer
from google.protobuf.timestamp_pb2 import Timestamp


def generate_request_id() -> str:
    return uuid.uuid4().hex


def get_metadata(request_id: Optional[str] = None) -> List:
    return [('request_id', request_id if request_id is not None else generate_request_id())]


def fill_flat_message(message: Message, entity: Mapping) -> None:
    for name, value in entity.items():
        if value is None:
            message.ClearField(name)
            continue

        message_attr = getattr(message, name)

        if isinstance(message_attr, Timestamp):
            message_attr.FromDatetime(value)
        else:
            setattr(message, name, value)


def message_to_dict(
    msg: Message,
    set_default_values: bool = True,
    use_integers_for_enums: bool = True,
    convert_datetime_to_utc: bool = True
) -> Dict[str, Any]:
    """
    The function is not tested to work with fields of google.protobuf.Struct etc.
    Function converts 'amount' field_descriptor from string to Decimal and Timestamp fields
    to datetime.
    """

    def pythonify_value(field_name: str, field_value) -> Any:
        if isinstance(field_value, Timestamp):
            dt = field_value.ToDatetime()
            if convert_datetime_to_utc:
                dt = dt.replace(tzinfo=timezone.utc)
            return dt
        if isinstance(field_value, str) and field_name == 'amount':
            return Decimal(field_value)

        return field_value

    def is_map_type_field(field_descriptor: FieldDescriptor) -> bool:
        if field_descriptor.label != FieldDescriptor.LABEL_REPEATED:
            return False
        if field_descriptor.message_type is None:
            return False
        if not field_descriptor.message_type.GetOptions().map_entry:
            return False

        return True

    def is_enum_type_field(field_descriptor: FieldDescriptor) -> bool:
        return field_descriptor.type == field_descriptor.TYPE_ENUM

    def get_enum_value_name(field_name: str, field_value: Any, local_msg: Message) -> str:
        return local_msg.DESCRIPTOR.fields_by_name[field_name].enum_type. \
            values_by_number.get(field_value).name

    def get_actual_field_value(field_name: str, local_msg: Message) -> Any:
        for field_descriptor, field_value in local_msg.ListFields():
            if field_descriptor.name == field_name:
                if isinstance(field_value, RepeatedCompositeContainer):
                    return [convert_descriptor(f) for f in field_value]
                if isinstance(field_value, RepeatedScalarContainer):
                    if is_enum_type_field(field_descriptor) and not use_integers_for_enums:
                        return [get_enum_value_name(field_name, n, local_msg) for n in field_value]
                    return list(field_value)
                if isinstance(field_value, MessageMapContainer):
                    return {k: convert_descriptor(field_value[k]) for k in field_value}
                if is_enum_type_field(field_descriptor) and not use_integers_for_enums:
                    return get_enum_value_name(field_name, field_value, local_msg)

                return pythonify_value(field_name, field_value)

        return None

    def convert_descriptor(local_msg: Message) -> Dict[str, Any]:
        result = {}
        if len(local_msg.DESCRIPTOR.oneofs_by_name.items()):
            result['which_one_of'] = {}
            for k, v in local_msg.DESCRIPTOR.oneofs_by_name.items():
                result['which_one_of'][k] = local_msg.WhichOneof(k)

        for field_name, field_descriptor in local_msg.DESCRIPTOR.fields_by_name.items():
            value = get_actual_field_value(field_name, local_msg)

            if value is None:
                if not set_default_values:
                    continue

                # Any composite type has a default value as empty list,
                # but we want a dict for the map
                if is_map_type_field(field_descriptor):
                    value = {}
                else:
                    value = field_descriptor.default_value

                # for case if enum value has been set to 0 (default value) or not specified
                if is_enum_type_field(field_descriptor) and not use_integers_for_enums:
                    value = get_enum_value_name(field_name, value, local_msg)

            result[field_name] = value

        return result

    return convert_descriptor(msg)


OptionalGrpcTimeout = Optional[Union[str, int, float]]


def correct_timeout(timeout: OptionalGrpcTimeout = None) -> Optional[float]:
    if timeout is not None:
        timeout = float(timeout)
        if timeout <= 0:
            timeout = None

    return timeout



from __future__ import annotations

from typing import Dict

import grpc

from .grpc_utils import OptionalGrpcTimeout, correct_timeout, get_metadata, message_to_dict
from .protos.exchange_info import api_pb2 as exchange_info_api
from .protos.exchange_info.api_pb2_grpc import ExchangeInfoApiStub


class ExchangeInfoClient:
    @classmethod
    def factory(cls, dsn: str, **kwargs) -> ExchangeInfoClient:
        """
        Args:
            kwargs: see keyword arguments in the constructor
        """
        if not dsn:
            raise ValueError(f'Parameter is not specified: dsn')

        channel = grpc.insecure_channel(dsn)

        return cls(channel, **kwargs)

    def __init__(
        self,
        channel: grpc.Channel,
        timeout_get_entities: OptionalGrpcTimeout = None
    ):
        """
        Args:
            timeout_get_entities: grpc timeout in sec.
        """
        self._channel = channel
        self._stub = ExchangeInfoApiStub(channel)
        self._timeout_get_entities = correct_timeout(timeout_get_entities)

    @property
    def timeout_get_entities(self):
        return self._timeout_get_entities

    def destroy(self) -> None:
        if getattr(self, '_channel', None):
            self._channel.close()

    def get_entities(self, request_id: str) -> Dict[str, Dict[int, Dict]]:
        request = exchange_info_api.EntityKinds()
        request.entity_kinds.extend([
            exchange_info_api.ASSET,
            exchange_info_api.SYMBOL,
            exchange_info_api.MARKET,
            exchange_info_api.EXCHANGE
        ])

        response = self._stub.GetEntities(
            request,
            metadata=get_metadata(request_id),
            timeout=self._timeout_get_entities
        )

        return message_to_dict(response)



import asyncio
import datetime
import json
from dataclasses import dataclass
from functools import partial
from multiprocessing import Event, Pipe, Process
from multiprocessing.connection import Connection
from typing import Any, Awaitable, Callable, Dict, List, Mapping, Optional, Type

import aio_pika

from . import DepthConsumeKey, MessageConsumerCollection, PipeRequest, PipeResponseRouter, \
    TransportFactory, TransportFactoryException
from .exchange_info_client import ExchangeInfoClient
from .grpc_utils import generate_request_id
from ..asyncio_helper import AsyncProgramEnv, run_program_forever
from ..tools import Singleton

AioPikaConsumeCallable = Callable[[aio_pika.IncomingMessage], Any]


@dataclass(frozen=True)
class InitExchangeEntitiesRequest(PipeRequest):
    dsn: str
    timeout: float


@dataclass(frozen=True)
class ConsumeDepthScrapingRequest(PipeRequest):
    dsn: str
    exchange: str
    consume_keys: frozenset


class RabbitConnection(
    metaclass=Singleton,
    singleton_hash_args=[0],
    singleton_hash_kwargs=['dsn']
):
    def __init__(self, dsn: str):
        self._dsn: str = str(dsn).strip()
        self._connection: Optional[aio_pika.Connection] = None

    @property
    def connection(self):
        return self._connection

    async def create_channel(self, prefetch_count: int = 100) -> aio_pika.Channel:
        if not self._connection:
            self._connection = await aio_pika.connect_robust(self._dsn)

        channel = await self._connection.channel()
        await channel.set_qos(prefetch_count=prefetch_count)

        return channel


class RabbitQueueConsumer:
    def __init__(
        self,
        channel: aio_pika.Channel,
        exchange_name: str,
        on_message: AioPikaConsumeCallable
    ):
        self._channel = channel
        self._exchange_name = str(exchange_name).strip()
        self._on_message = on_message
        self._exchange: Optional[aio_pika.Exchange] = None
        self._queue: Optional[aio_pika.Queue] = None

    @property
    def exchange_name(self):
        return self._exchange_name

    @property
    def channel(self):
        return self._channel

    @property
    def exchange(self):
        return self._exchange

    @property
    def queue(self):
        return self._queue

    async def create_queue(self) -> aio_pika.Queue:
        self._exchange = await self.channel.declare_exchange(self._exchange_name, passive=True)
        self._queue = await self.channel.declare_queue(exclusive=True)

        await self._queue.consume(self._on_message, no_ack=True)

        return self._queue


class RealTransportFactory(TransportFactory):
    def __init__(self):
        super().__init__()
        self._process: Optional[Process] = None
        self._process_ready = Event()
        self._parent_connection, self._child_connection = Pipe()
        self._response_router: Optional[PipeResponseRouter] = None
        self._exchange_info_dsn: Optional[str] = None
        self._exchange_info_get_entities_timeout: float = 5.0
        self._depth_scraping_queue_dsn: Optional[str] = None
        self._depth_scraping_queue_exchange: Optional[str] = None

    def configure_endpoints(
        self,
        exchange_info_dsn: Optional[str] = None,
        exchange_info_get_entities_timeout: Optional[float] = None,
        depth_scraping_queue_dsn: Optional[str] = None,
        depth_scraping_queue_exchange: Optional[str] = None,
    ) -> None:
        def sanity_string(value: Optional[str]) -> Optional[str]:
            if value is not None:
                return str(value).strip()

        self._exchange_info_dsn = sanity_string(exchange_info_dsn)
        if exchange_info_get_entities_timeout is not None:
            self._exchange_info_get_entities_timeout = float(exchange_info_get_entities_timeout)
        self._depth_scraping_queue_dsn = sanity_string(depth_scraping_queue_dsn)
        self._depth_scraping_queue_exchange = sanity_string(depth_scraping_queue_exchange)

    async def init(self, loop_debug: Optional[bool] = None) -> None:
        if self._process:
            raise RuntimeError('A process for RealTransportFactory should be created only once')

        self._process = RealTransportProcess(
            loop_debug=loop_debug,
            ready_event=self._process_ready,
            connection=self._child_connection
        )
        self._process.start()

        self._response_router = PipeResponseRouter(self._parent_connection)
        task = asyncio.create_task(self._response_router.start())

        def task_done_cb(t: asyncio.Task) -> None:
            if t.cancelled():
                self.shutdown()
                return

            if t.exception():
                self.shutdown()
                raise t.exception()

        task.add_done_callback(task_done_cb)

        if not self._process_ready.wait(2):
            raise RuntimeError('Failed to initialize RealTransportFactory in time')

    def shutdown(self) -> None:
        if self._process:
            self._process.terminate()

    async def init_exchange_entities(
        self,
        on_response: Callable[..., Awaitable]
    ) -> MessageConsumerCollection:
        request = InitExchangeEntitiesRequest(
            self._exchange_info_dsn,
            self._exchange_info_get_entities_timeout
        )
        result = self._response_router.prepare_consumers_of_response(request)
        result.add_consumer(on_response)
        self._parent_connection.send(request)

        return result

    async def get_depth_scraping_consumer(
        self,
        on_response: Callable[..., Awaitable],
        consume_keys: Optional[List[DepthConsumeKey]] = None
    ) -> MessageConsumerCollection:
        request = ConsumeDepthScrapingRequest(
            self._depth_scraping_queue_dsn,
            self._depth_scraping_queue_exchange,
            frozenset(consume_keys)
        )
        result = self._response_router.prepare_consumers_of_response(request)
        result.add_consumer(on_response)
        self._parent_connection.send(request)

        return result


class RealTransportProcess(Process):
    def __init__(
        self,
        *args,
        loop_debug: Optional[bool] = None,
        ready_event: Event,
        connection: Connection,
        poll_delay: float = 0.001,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._loop_debug = loop_debug
        self._ready_event = ready_event
        self._connection = connection
        self._poll_delay = float(poll_delay)
        self._handlers: Mapping[Type[PipeRequest], Callable[..., Awaitable]] = {
            InitExchangeEntitiesRequest: self.init_exchange_entities,
            ConsumeDepthScrapingRequest: self.consume_depth_scraping,
        }

    def run(self) -> None:
        run_program_forever(self.main, loop_debug=self._loop_debug)

    async def main(self, program_env: AsyncProgramEnv) -> None:
        def exception_handler(local_loop: asyncio.AbstractEventLoop, context: Dict) -> None:
            if 'exception' in context:
                self._notify_owner_process(context['exception'])

        program_env.exception_handler_patch = exception_handler

        self._ready_event.set()

        while True:
            if not self._connection.poll():
                await asyncio.sleep(self._poll_delay)
                continue

            request = self._connection.recv()

            handler = self._find_handler_for_request(request)
            asyncio.create_task(handler(request))

    async def init_exchange_entities(self, request: InitExchangeEntitiesRequest) -> None:
        client = ExchangeInfoClient.factory(request.dsn, timeout_get_entities=request.timeout)
        entities = client.get_entities(generate_request_id())
        self._respond_to_owner_request(request, entities)

        client.destroy()

    async def consume_depth_scraping(self, request: ConsumeDepthScrapingRequest) -> None:
        connection = RabbitConnection(request.dsn)
        channel = await connection.create_channel(100)
        cb = partial(self._depth_scraping_callback, request)
        consumer = RabbitQueueConsumer(channel, request.exchange, cb)
        queue = await consumer.create_queue()

        if not request.consume_keys:
            routing_keys = ['#']
        else:
            routing_keys = [k.format_for_rabbitmq() for k in request.consume_keys]

        for routing_key in routing_keys:
            await queue.bind(consumer.exchange, routing_key)

    def _notify_owner_process(self, original_exception: Exception) -> None:
        # This wrapping is required to don't fire unnecessary errors about serialization
        # of the exception. Otherwise a framework user will see unrequired spam about
        # pickling RLock etc in logs.
        wrapped_exception = TransportFactoryException('An error in the transport process')
        wrapped_exception.__cause__ = original_exception

        self._connection.send([wrapped_exception])

    def _find_handler_for_request(self, request: PipeRequest) -> Callable[..., Awaitable]:
        request_type = type(request)
        if request_type not in self._handlers:
            raise ValueError(f'No handler for request type {request_type}')

        return self._handlers[request_type]

    def _respond_to_owner_request(self, request: PipeRequest, content: Any):
        self._connection.send([request, content])

    def _depth_scraping_callback(
        self,
        request: ConsumeDepthScrapingRequest,
        message: aio_pika.IncomingMessage
    ) -> None:
        body = json.loads(message.body)
        args = [
            body['exchange'],
            body['market'],
            body['symbol'],
            datetime.datetime.fromisoformat(body['now']),
            body['depth']['bids'],
            body['depth']['asks'],
        ]

        self._respond_to_owner_request(request, args)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protos/common.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='protos/common.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x13protos/common.proto\"\x1c\n\nBoolResult\x12\x0e\n\x06result\x18\x01 \x01(\x08\",\n\x08MarketId\x12\x10\n\x08\x65xchange\x18\x01 \x01(\t\x12\x0e\n\x06symbol\x18\x02 \x01(\t\"#\n\x0bSessionHash\x12\x14\n\x0csession_hash\x18\x01 \x01(\tb\x06proto3')
)




_BOOLRESULT = _descriptor.Descriptor(
  name='BoolResult',
  full_name='BoolResult',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='result', full_name='BoolResult.result', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=23,
  serialized_end=51,
)


_MARKETID = _descriptor.Descriptor(
  name='MarketId',
  full_name='MarketId',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='exchange', full_name='MarketId.exchange', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='symbol', full_name='MarketId.symbol', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=53,
  serialized_end=97,
)


_SESSIONHASH = _descriptor.Descriptor(
  name='SessionHash',
  full_name='SessionHash',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='session_hash', full_name='SessionHash.session_hash', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=99,
  serialized_end=134,
)

DESCRIPTOR.message_types_by_name['BoolResult'] = _BOOLRESULT
DESCRIPTOR.message_types_by_name['MarketId'] = _MARKETID
DESCRIPTOR.message_types_by_name['SessionHash'] = _SESSIONHASH
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

BoolResult = _reflection.GeneratedProtocolMessageType('BoolResult', (_message.Message,), {
  'DESCRIPTOR' : _BOOLRESULT,
  '__module__' : 'protos.common_pb2'
  # @@protoc_insertion_point(class_scope:BoolResult)
  })
_sym_db.RegisterMessage(BoolResult)

MarketId = _reflection.GeneratedProtocolMessageType('MarketId', (_message.Message,), {
  'DESCRIPTOR' : _MARKETID,
  '__module__' : 'protos.common_pb2'
  # @@protoc_insertion_point(class_scope:MarketId)
  })
_sym_db.RegisterMessage(MarketId)

SessionHash = _reflection.GeneratedProtocolMessageType('SessionHash', (_message.Message,), {
  'DESCRIPTOR' : _SESSIONHASH,
  '__module__' : 'protos.common_pb2'
  # @@protoc_insertion_point(class_scope:SessionHash)
  })
_sym_db.RegisterMessage(SessionHash)


# @@protoc_insertion_point(module_scope)



import datetime
from typing import Optional

from .transport import TransportFactory


class Asset:
    def __init__(
        self,
        transport_factory: TransportFactory,
        id: int,
        tag: str,
        name: str,
        precision: int,
        create_time: datetime.datetime,
        delete_time: Optional[datetime.datetime]
    ):
        self._transport_factory = transport_factory
        self._id: int = int(id)
        self._tag: str = str(tag).strip()
        self._name: str = str(name).strip()
        self._precision: int = int(precision)
        self._create_time: datetime.datetime = create_time
        self._delete_time: Optional[datetime.datetime] = delete_time


class Symbol:
    def __init__(
        self,
        transport_factory: TransportFactory,
        id: int,
        base_asset_id: int,
        quote_asset_id: int,
        create_time: datetime.datetime,
        delete_time: Optional[datetime.datetime]
    ):
        self._transport_factory = transport_factory
        self._id: int = int(id)
        self._base_asset_id: int = int(base_asset_id)
        self._quote_asset_id: int = int(quote_asset_id)
        self._create_time: datetime.datetime = create_time
        self._delete_time: Optional[datetime.datetime] = delete_time



import asyncio
import os
import signal
from functools import partial
from typing import Any, Awaitable, Callable, Dict, List, Optional, Sequence

from .structlogger import get_logger

logger = get_logger('galts_trade_api')

LoopExceptionHandlerCallable = Callable[[asyncio.AbstractEventLoop, Dict[str, Any]], Any]


def signal_handler(sig: signal.Signals, loop: asyncio.AbstractEventLoop) -> None:
    logger.info('Received exit signal', process_id=os.getpid(), signal=sig.name)
    loop.create_task(shutdown(loop))


# Inspired by https://www.roguelynn.com/words/asyncio-exception-handling/
async def shutdown(loop: asyncio.AbstractEventLoop) -> None:
    """Cleanup tasks tied to the program's shutdown."""
    logger.info(f'Shutting down', process_id=os.getpid())

    other_tasks = [t for t in asyncio.all_tasks(loop) if t is not asyncio.current_task(loop)]
    logger.debug('Cancelling outstanding tasks', process_id=os.getpid(), amount=len(other_tasks))
    await _cancel_tasks(loop, other_tasks)

    loop.stop()


# Do it like asyncio.run()
async def _cancel_tasks(loop: asyncio.AbstractEventLoop, tasks: List) -> None:
    if not tasks:
        return

    for task in tasks:
        task.cancel()

    await asyncio.gather(*tasks, return_exceptions=True, loop=loop)

    for task in tasks:
        if task.cancelled():
            continue

        if task.exception():
            loop.default_exception_handler({
                'message': 'Unhandled exception during shutdown',
                'exception': task.exception(),
                'task': task,
            })


def run_program_forever(
    target: Callable[..., Awaitable],
    loop: Optional[asyncio.AbstractEventLoop] = None,
    loop_debug: Optional[bool] = None,
    handle_signals: Optional[Sequence[signal.Signals]] = None
) -> None:
    """
    Args:
        handle_signals: If None then these signals will be handled: SIGTERM, SIGINT.
    """
    if not loop:
        loop = asyncio.new_event_loop()
        new_loop_created = True
    else:
        new_loop_created = False

    if loop_debug is not None:
        loop.set_debug(loop_debug)

    if handle_signals is None:
        handle_signals = {signal.SIGTERM, signal.SIGINT}

    for sig in handle_signals:
        loop.add_signal_handler(sig, partial(signal_handler, sig, loop))

    try:
        env = AsyncProgramEnv()
        loop.set_exception_handler(env.exception_handler)
        loop.create_task(target(env))
        loop.run_forever()
    finally:
        if new_loop_created:
            loop.run_until_complete(loop.shutdown_asyncgens())
            loop.close()

        logger.info('Successfully shutdown', process_id=os.getpid())


class AsyncProgramEnv:
    """
    The main purpose of the class is to provide an ability to extend the event loop exception
    handler. A user can perform additional shutdown tasks in the extension.
    """

    def __init__(self):
        self._exception_handler_patch: Optional[LoopExceptionHandlerCallable] = None

    @property
    def exception_handler_patch(self):
        return self._exception_handler_patch

    @exception_handler_patch.setter
    def exception_handler_patch(self, value: Optional[LoopExceptionHandlerCallable] = None):
        if value is not None and not callable(value):
            raise TypeError('Value should be a callable or None')

        self._exception_handler_patch = value

    def exception_handler(self, loop: asyncio.AbstractEventLoop, context: Dict[str, Any]) -> None:
        if self.exception_handler_patch:
            self.exception_handler_patch(loop, context)

        logger.info('Caught exception', process_id=os.getpid())
        loop.default_exception_handler(context)

        loop.create_task(shutdown(loop))
